package be.cylab.java.wowa.training;

import org.datavec.api.records.reader.RecordReader;
import org.datavec.api.records.reader.impl.csv.CSVRecordReader;
import org.datavec.api.split.FileSplit;
import org.datavec.api.util.ClassPathResource;
import org.deeplearning4j.datasets.datavec.RecordReaderDataSetIterator;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.eval.ROC;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.SplitTestAndTrain;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.preprocessor.DataNormalization;
import org.nd4j.linalg.dataset.api.preprocessor.NormalizerStandardize;
import org.nd4j.linalg.lossfunctions.LossFunctions;

import java.io.IOException;

/**
 * Class to test neural network learning.
 */

public final class LearningNeuralNetwork {

    /**
     * Class count.
     */
    public static final int CLASSES_COUNT = 2;
    /**
     * Features count.
     */
    public static final int FEATURES_COUNT = 5;


    private DataSet training_data;
    private DataSet testing_data;
    private double learning_rate;
    private OptimizationAlgorithm algorithm;
    private Activation activation_function;


    /**
     * Constructor.
     *
     * @param file_name
     */
    public LearningNeuralNetwork(final String file_name,
                                 final double learning_rate,
                                 final OptimizationAlgorithm algorithm,
                                 final Activation activation_function
    ) {
        try (RecordReader record_reader = new CSVRecordReader(0, ',')) {
            record_reader.initialize(new FileSplit(
                    new ClassPathResource(file_name).getFile()
            ));
            DataSetIterator iterator = new RecordReaderDataSetIterator(
                    record_reader, 12468, FEATURES_COUNT, CLASSES_COUNT);
            DataSet all_data = iterator.next();
            all_data.shuffle();
            DataNormalization normalizer = new NormalizerStandardize();
            normalizer.fit(all_data);
            normalizer.transform(all_data);

            SplitTestAndTrain test_and_train = all_data.splitTestAndTrain(0.90);
            this.training_data = test_and_train.getTrain();
            this.testing_data = test_and_train.getTest();

        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        this.learning_rate = learning_rate;
        this.algorithm = algorithm;
        this.activation_function = activation_function;
    }

    /**
     * Method to build a simple neural network, train and test it.
     *
     * @param neurons_number
     */
    public void learning(final int neurons_number) {

        MultiLayerConfiguration configuration
                = new NeuralNetConfiguration.Builder()
                .iterations(1000)
                .activation(this.activation_function)
                .optimizationAlgo(
                        this.algorithm)
                .weightInit(WeightInit.XAVIER)
                .learningRate(this.learning_rate)
                .regularization(true).l2(0.0001)
                .list()
                .layer(0, new DenseLayer.Builder().nIn(FEATURES_COUNT)
                        .nOut(neurons_number).build())
                .layer(1, new DenseLayer.Builder().nIn(neurons_number)
                        .nOut(neurons_number).build())
                .layer(2, new OutputLayer.Builder(LossFunctions
                        .LossFunction.NEGATIVELOGLIKELIHOOD)
                        .activation(Activation.SOFTMAX)
                        .nIn(neurons_number)
                        .nOut(CLASSES_COUNT).build())
                .backprop(true).pretrain(false)
                .build();

        MultiLayerNetwork model = new MultiLayerNetwork(configuration);
        model.init();
        model.fit(this.training_data);

        INDArray output
                = model.output((this.testing_data.getFeatureMatrix()));

        Evaluation eval = new Evaluation(CLASSES_COUNT);
        eval.eval(this.testing_data.getLabels(), output);
        System.out.println(eval.stats());
        ROC roc = new ROC(CLASSES_COUNT);
        roc.eval(this.testing_data.getLabels(), output);
        System.out.println(roc.stats());
    }
}
